#include "rs232.h"
#define _VOID_FINALIZE
#include "voidbind.h"
#include "gcc2catalina.h" // startcog

// Operate a synthetic rs232 port in full duplex
// This is just a generic io port.

static
void destroy( rs232_ct* self ) {
	if (self && self->m.cog ) cogstop( self->m.cog - 1 );
	self->m.cog = 0;
}

static 
void init(rs232_ct* self,
	   	int txpin, int rxpin, int mode, int baudrate
	){

	long *i;
	long bitQtime;  // roughly 1/3 bit time for metastable flip-flop recovery
	long bitQQtime; // fixup for roundoff of bit time

	destroy( self ); // Make sure old cog, if any, is deactivated

	bitQQtime = CLKFREQ;
	bitQQtime = (bitQQtime+(baudrate>>1)) / baudrate; // Compute 
	bitQtime  = bitQQtime/3; // compute 1/3 bit time
	bitQQtime -= (bitQtime<<1) + bitQtime; // compute residual.
	if (bitQQtime!=2) bitQQtime+=bitQtime; // Do a minimum error fixup
	else { bitQQtime=bitQtime; ++bitQtime; }

	// I need to initialize the data in the RS232 assembly language driver.
	// Most of this information does not belong in the struct, but as I
	// ned a place to pass these pieces of information -- I'm going to
	// temporarily overwrite the struct with the initialization data, 
	// as if it were  merely an array of longs. (Don't have to make temp ram)
	// Then the assembly language routine will reinitialize the struct properly
	// by erasing any un-needed data.

	i=(long*)(void*)self;
	*i   = (long)(void*)self->m.tx_buffer;
	*++i = (long)(void*)self->m.rx_buffer;
	*++i = (long)(void*)self->m.pTxe;
	*++i = (long)(void*)self->m.pRxe;
	*++i = bitQtime;
	*++i = bitQQtime;
	*++i = (1L << txpin);
	*++i = (1L << rxpin);
	*++i = (long)mode; // Not used, yet.
	self->m.sIn=self->m.sOut=self->m.s_buffer;

	// Fire in the hole!

	self->m.cog = cognew( rs232_S, i ) + 1;
}

// Non blocking reads, returns -1 if no data.  Character otherwise.

static
int nbGetchar( rs232_ct* self ) {
	volatile char *p = self->m.pRx;
	char ch = *p;
	if (!ch) return -1; // Empty and idle
	*p=0;	
	if (++p == self->m.pRxe) p-=SZ_RX;
	self->m.pRx=p;
	return ch;
}

static
int _getchar( rs232_ct* self ) {
	int ch;
	do ch=nbGetchar(self); while(ch<=0);
	return ch;
}

static 
void iFlush( rs232_ct* self ) {
	while( nbGetchar(self) ) continue;
}

static
int nbPutchar( rs232_ct* self, unsigned char ch ) {
	volatile char* p = self->m.pTx; // Present location in buffer
	if (*(p)) return -1; // if char at present, buffer Full & busy.
	*p = ch;
	if (++p == self->m.pTxe) p-=SZ_TX;
	self->m.pTx=p;
	return ch;
}

static
int _putchar( rs232_ct* self, unsigned char ch ) {
	while( nbPutchar( self, ch )<=0 ) continue;
	return ch;
}

// Non blocking put string
static
int nbPuts( rs232_ct* self, unsigned char* str ) {
	unsigned char* p = str-1;
	while (*(++p) && nbPutchar(self,*p) ) continue;
	if (*p) return (int)(str-p)-1; // Return negative char count -1 on error.
	return 1;
}

// Blocking put string
static
int _puts( rs232_ct* self, unsigned char* str ) {
	--str;
	while (*(++str)) _putchar(self, *str );
	return 1;
}

// Slow putstring
// A large buffer, but only injects three characters into transmitter
// per call, max.
// Must be polled, preferably with str=NULL, or empty string, to finish
// emptying larger buffer.

// If the buffer attempts to over-fill, function will block until enough 
// data is transmitted for buffer to hold remaining data.

static
void sputs( rs232_ct* self, unsigned char* str ) {
	char ch;
	volatile char *p; // track up to 3 bytes of tx data
    char *q; // enquing pointer

	// If buffer not empty, transmit some of the data
	while (ch = *(self->m.sOut)) { // if data to transmit
		p = self->m.pTx-3;
		if (p < self->m.tx_buffer ) p += SZ_TX;
		if (*p) break; // 3 or more chars available, so don't stuff more.
		nbPutchar( self, ch ); 
		*(self->m.sOut)=0;
		if (++(self->m.sOut)==self->m.sEnd) self->m.sOut=self->m.s_buffer;
	}
	if (!str) return;

	// Copy entire string into buffer, block if not enough space.

	q=self->m.sIn;	
	while ( ch=*str ) {
		while (*q) sputs( self, NULL ); // Wait to drain.
		*q=ch; // copy in a new char
		if (++q == self->m.sEnd) q=self->m.s_buffer;
		++str;
	}
	self->m.sIn=q;

	return;
}

static
void sputchar( rs232_ct* self, unsigned char ch ){
	 sputs(self,(unsigned char[2]){ ch,0 } );
	 return;
}

// linker information for the library
rs232_mt rs232_mt_ld = {
	.init=&init,
	.destroy=&destroy,
	.nbGetchar=&nbGetchar,
	.getchar=&_getchar,
	.iFlush=&iFlush,
	.nbPutchar=&nbPutchar,
	.putchar=&_putchar,
	.nbPuts=&nbPuts,
	.puts=&_puts,
	.sputs=&sputs,
	.sputchar=&sputchar,
};

